;; ------------------------------------------------------------------
;; Basic ball operations: see pingp-lib.ss
;; ---------------------------------------

;; More primitives on posn:
;; ------------------------
;; posn+: posn posn -> posn (component-wise addition)
(define (posn+ p1 p2)
  (make-posn (+ (posn-x p1) (posn-x p2)) (+ (posn-y p1) (posn-y p2))))

;; posn+: number posn -> posn (component-wise muliplication)
(define (posn*s f p)
  (make-posn (* f (posn-x p)) (* f (posn-y p))))

;; balls = (cons posn (cons posn null))
;;   the first posn is the position, 
;;   the second one the speed per time unit

;; The ball representation and some abasic primitives:
;; ---------------------------------------------------
(define (make-ball pos speed) (cons pos (cons speed null)))
(define (ball-posn ball) (first ball))
(define (ball-speed ball) (first (rest ball))) 

;; ns-direction : ball => {'NORTH, 'SOUTH, 'none}
(define (ns-direction ball)
  (cond
    ((< (posn-y (ball-speed ball)) 0) 'NORTH)
    ((> (posn-y (ball-speed ball)) 0) 'SOUTH)))

;; ns-dist-to-wall : ball -> number (dist to n/s wall moving towards)
(define (ns-dist-to-wall ball)
  (cond
    ((eq? (ns-direction ball) 'NORTH) (- (posn-y (ball-posn ball)) NORTH))
    ((eq? (ns-direction ball) 'SOUTH) (- SOUTH (posn-y (ball-posn ball))))))

;; dist-to-paddle : ball -> number (dist to paddle)
(define (dist-to-paddle ball)
  (- PADDLE-X (posn-x (ball-posn ball))))

;; ns-time-to-wall : ball -> number (time before ns wall is hit)
(define (ns-time-to-wall ball)
  (/ (ns-dist-to-wall ball)
     (abs (posn-y (ball-speed ball)))))

;; time-to-paddle : ball -> number (time before paddle is hit)
(define (time-to-paddle ball)
  (/ (dist-to-paddle ball)
     (abs (posn-x (ball-speed ball)))))

;; Moving a Ball
;; -------------
;; move-in-box : ball number -> ball or #f (if the ball gets destroyed)
(define (move-in-box ball t)
  (cond
    ((eq? (bounces-off ball t) 'NORTH-SOUTH)
     (move-in-box
       (ns-bounce
	 (move-ball ball (ns-time-to-wall ball)))
       (- t (ns-time-to-wall ball))))
    ((eq? (bounces-off ball t) 'PADDLE) #f)
    (else (move-ball ball t))))

;; bounces-off : ball number -> {'NORTH-SOUTH,'PADDLE,'none}
(define (bounces-off ball t)
  (cond
    ((and (<= (ns-time-to-wall ball) (time-to-paddle ball))
	  (<= (ns-time-to-wall ball) t))
     'NORTH-SOUTH)
    ((and (<= 0 (time-to-paddle ball) (ns-time-to-wall ball))
	  (<= (time-to-paddle ball) t))
     (cond 
       ((landed-on-paddle? (ball-posn (move-ball ball (time-to-paddle ball))))
	'PADDLE)
       (else 'none)))
    (else 'none)))

;; move : ball number -> ball
(define (move-ball ball t)
  (make-ball (posn+ (ball-posn ball) (posn*s t (ball-speed ball)))
	     (ball-speed ball)))

;; ns-bounce : ball -> ball
(define (ns-bounce ball)
  (make-ball (ball-posn ball)
	     (make-posn (posn-x (ball-speed ball)) 
			(- (posn-y (ball-speed ball))))))

;; ------------------------------------------------------------------
;; Dealing with collections of balls
;; ---------------------------------

;; mk-balls : natnum -> list-of-balls
(define (mk-balls a-nn)
  (cond
    ((zero? a-nn) null)
    (else
     (cons 
      (make-ball (make-posn (random-between FAR_WEST EAST) (random SOUTH))
		 (make-posn (random-between MIN-X-SPEED MAX-X-SPEED) 
			    (random-between MIN-Y-SPEED MAX-Y-SPEED)))
      (mk-balls (- a-nn 1))))))

;; random-between : int int -> int (randomly in betweeen he two inputs)
(define (random-between low high)
  (+ low (random (+ (abs low) (abs high)))))

;; move-balls : list-of-balls -> list-of-balls 
(define (move-balls loballs)
  (cond
    ((null? loballs) null)
    (else (cons (move-in-box (first loballs) 1) (move-balls (rest loballs))))))

;; remove-balls-hit-paddle : list-of-balls/#f -> list-of-balls 
;;  (those that hit paddle during a move or are outside after a move)
(define (remove-balls-hit-paddle loballs)
  (cond
    ((null? loballs) null)
    (else (cond
	    ((boolean? (first loballs))
	     (remove-balls-hit-paddle (rest loballs)))
	    (else
	     (cons (first loballs) (remove-balls-hit-paddle (rest loballs))))))))

;; remove-outside-balls : list-of-balls -> list-of-balls 
;;  (those that hit paddle during a move or are outside after a move)
(define (remove-outside-balls loballs)
  (cond
    ((null? loballs) null)
    (else (cond
	    ((inside? (first loballs))
	     (cons (first loballs) (remove-outside-balls (rest loballs))))
	    (else
	     (remove-outside-balls (rest loballs)))))))

;; inside? : ball -> boolean (is the ball inside the user-defined space)
(define (inside? aball)
  (and (<= FAR_WEST (posn-x (ball-posn aball)) EAST)
       (<= NORTH    (posn-y (ball-posn aball)) SOUTH)))

;; balls-posn : list-of-balls -> list-of-posn (a projection)
(define (balls-posn l)
  (cond
    ((null? l) null)
    (else (cons (ball-posn (first l)) (balls-posn (rest l))))))

;; the true extent of the space
(define FAR_WEST (* -1 EAST))

(define MIN-X-SPEED 05)
(define MAX-X-SPEED 15)
(define MIN-Y-SPEED 1)
(define MAX-Y-SPEED 4)

